import re

from django.contrib import admin
from django import forms

from . import models
from . import views


class UrlMappingAdmin(admin.ModelAdmin):
    list_display = ['source_url', 'target_url', 'expire_after', 'last_usage']
    search_fields = ['source_url', 'target_url']


class UrlRegexpMappingForm(forms.ModelForm):
    def clean(self):
        cleaned_data = super().clean()
        expected = cleaned_data.get('test_target_url', '')
        try:
            result = views.generate_url(
                self.cleaned_data.get('source_mapping', ''),
                self.cleaned_data.get('target_mapping', ''),
                cleaned_data.get('test_source_url', '')
            )
        except re.error as e:
            self.add_error('source_mapping', 'One of regexps is invalid: %s' % str(e))
        else:
            if result != expected:
                self.add_error('test_source_url', 'Expected %s got %s' % (expected, result))
        return cleaned_data

    class Meta:
        model = models.UrlRegexpMapping
        fields = ['source_mapping', 'target_mapping', 'expire_after', 'mapping_duration',
                  'test_source_url', 'test_target_url']


class UrlRegexpMappingAdmin(admin.ModelAdmin):
    form = UrlRegexpMappingForm
    list_display = ['source_mapping', 'target_mapping', 'expire_after', 'mapping_duration', 'last_usage']
    search_fields = ['source_mapping', 'target_mapping']
    fieldsets = (
        (None, {
            'fields': ('source_mapping', 'target_mapping', 'expire_after', 'mapping_duration'),
        }),
        ('Test', {
            'fields': ('test_source_url', 'test_target_url'),
        }),
    )


admin.site.register(models.UrlMapping, UrlMappingAdmin)
admin.site.register(models.UrlRegexpMapping, UrlRegexpMappingAdmin)



from datetime import timedelta

from django.conf import settings
from django.db import models


class LastUsageLog(models.Model):
    used_date = models.DateTimeField()
    referer = models.TextField()
    user_agent = models.TextField()
    ip = models.GenericIPAddressField()
    user = models.ForeignKey('auth.User', null=True, on_delete=models.SET_NULL)

    def __str__(self):
        return str(self.used_date)


class UrlMapping(models.Model):
    source_url = models.CharField(max_length=200, unique=True)
    target_url = models.CharField(max_length=200)
    expire_after = models.DurationField(
        help_text='Time of inactivity after which the mapping expires. Format: "DD HH:MM:SS"',
        default=timedelta(days=180)
    )
    last_usage = models.OneToOneField(
        LastUsageLog, null=True, editable=False, on_delete=models.SET_NULL)

    def lock(self):
        return UrlMapping.objects.select_for_update().get(pk=self.pk)

    def __str__(self):
        return '%s -> %s' % (self.source_url, self.target_url)


class UrlRegexpMapping(models.Model):
    source_mapping = models.CharField(max_length=200, unique=True)
    target_mapping = models.CharField(max_length=200)
    expire_after = models.DurationField(
        help_text='Time of inactivity after which the mapping expires. Format: "DD HH:MM:SS"',
        default=timedelta(days=180))
    mapping_duration = models.DurationField(
        help_text='Time of inactivity after which the created url mapping expires. Format: "DD HH:MM:SS"',
        default=timedelta(days=180)
    )
    last_usage = models.OneToOneField(
        LastUsageLog, null=True, editable=False, on_delete=models.SET_NULL)
    test_source_url = models.CharField(max_length=200)
    if settings.DEBUG:
        test_source_url.help_text = (
            '<b><span style="color: red;">'
            'You are in the debug mode - url mapping is disabled. '
            'You can test it here only.'
            '</span></b>'
        )
    test_target_url = models.CharField(
        max_length=200, help_text=('Provide example old/new urls you are attempting to map. Test urls will be used to test your regexp mappings.'
                                   '<br>For example <b>/a/(?P&lt;pk&gt;[0-9]+)/</b> mapping with <b>/b/\g&lt;pk&gt;/</b> target mapping'
                                   ' will change /a/1/ to /b/1/.'))

    def __str__(self):
        return '%s -> %s' % (self.source_mapping, self.target_mapping)

    def lock(self):
        return UrlRegexpMapping.objects.select_for_update().get(pk=self.pk)


class RegexpGeneratedMapping(UrlMapping):
    regexp = models.ForeignKey(UrlRegexpMapping, on_delete=models.SET_NULL, null=True)



from unittest import mock

from django import test

from .. import models
from .. import views
from . import factories


class TestUrlResolverIntegration(test.TestCase):
    def test_if_url_is_resolved(self):
        factories.UrlMappingFactory(source_url='/bar', target_url='/foo')
        response = self.client.get('/bar')
        self.assertEqual('/foo', response.url)
        self.assertEqual(301, response.status_code)

    def test_if_url_is_resolved_by_regexp(self):
        factories.UrlRegexpMappingFactory(
            source_mapping=r'/profil-v/(?P<pk>[0-9]+)/',
            target_mapping=r'/profil-miejsca/\1/'
        )
        response = self.client.get('/profil-v/10/')
        self.assertEqual('/profil-miejsca/10/', response.url)
        self.assertEqual(301, response.status_code)

    def test_if_404_is_returned(self):
        response = self.client.get('/abc/10/')
        self.assertEqual(404, response.status_code)

    @mock.patch('url_migration.views.logger')
    def test_if_404_is_returned_for_invalid_redirect(self, logger):
        factories.UrlMappingFactory(
            source_url='/a/',
            target_url='/a/',
        )
        response = self.client.get('/a/')
        self.assertTrue(logger.exception.called)
        self.assertEqual(404, response.status_code)


class TestRegexpMapping(test.TestCase):
    META = {
        'HTTP_REFERER': 'http://google.pl',
        'HTTP_USER_AGENT': 'Firefox 1',
        'HTTP_X_REAL_IP': '10.0.0.2',
    }

    def test_if_get_or_create_url_mapping_returns_existing_mapping(self):
        rule = factories.UrlRegexpMappingFactory(
            source_mapping=r'/a/b/(?P<pk>[0-9]+)/')
        mapping = factories.RegexpGeneratedMappingFactory(
            source_url='/a/b/1/', regexp=rule)
        other_rule = factories.UrlRegexpMappingFactory(
            source_mapping=r'/a/(.*)/')

        request = mock.Mock(path='/a/b/1/')
        resolver = views.UrlResolver(request)
        result = resolver._get_or_create_url_mapping(other_rule)
        self.assertEqual(mapping, result)
        self.assertEqual(1, models.RegexpGeneratedMapping.objects.count())

    def test_if_get_or_create_url_mapping_creates_new_mapping(self):
        rule = factories.UrlRegexpMappingFactory(
            source_mapping=r'/a/(?P<pk>[0-9]+)/',
            target_mapping=r'/b/\1/',
        )
        factories.RegexpGeneratedMappingFactory(
            source_url='/a/1234/', regexp=rule)

        request = mock.Mock(path='/a/1/')
        resolver = views.UrlResolver(request)
        result = resolver._get_or_create_url_mapping(rule)
        expected = models.RegexpGeneratedMapping.objects.get(
            source_url='/a/1/',
            target_url='/b/1/',
            regexp=rule
        )
        self.assertEqual(expected, result)
        self.assertEqual(2, models.RegexpGeneratedMapping.objects.count())

    def test_if_exception_is_raised_for_redirect_loop(self):
        factories.UrlMappingFactory(
            source_url='/a/',
            target_url='/a/',
        )
        request = test.RequestFactory().get('/a/')
        resolver = views.UrlResolver(request)
        with self.assertRaises(views.InvalidRedirect):
            list(resolver.map())



import logging
import re

from django.db import transaction
from django import http
from django.utils import timezone
from django.views import defaults

from . import models


logger = logging.getLogger(__name__)


def page_not_found(request, template_name='404.html'):
    return next(get_response(request, template_name, UrlResolver))


def get_response(request, template_name, resolver_class):
    try:
        resolver = resolver_class(request)
        yield from resolver.map()
    except InvalidRedirect as e:
        logger.exception(e)
    yield defaults.page_not_found(request, template_name)


class UrlResolver:
    def __init__(self, request):
        self.request = request
        self.path = request.path
        self.meta = request.META

    def map(self):
        yield from self._map_by_path()
        yield from self._map_by_regexp()

    def _map_by_path(self):
        try:
            mapping = self._get_matching_url_mapping()
            yield self._log_and_get_response(mapping)
        except models.UrlMapping.DoesNotExist:
            return

    def _get_matching_url_mapping(self):
        return models.UrlMapping.objects.get(source_url=self.path)

    def _map_by_regexp(self):
        for mapping in self._get_matching_regexp_mappings():
            url_mapping = self._get_or_create_url_mapping(mapping)
            self._update_usage(url_mapping.regexp)
            yield self._log_and_get_response(url_mapping)

    def _log_and_get_response(self, rule):
        path = self.request.get_full_path()
        redirect_to = path.replace(rule.source_url, rule.target_url)
        if path == redirect_to:
            raise InvalidRedirect('Redirect loop detected for {}'.format(path))
        else:
            self._update_usage(rule)
            return self._redirect_response(redirect_to)

    def _get_matching_regexp_mappings(self):
        for mapping in models.UrlRegexpMapping.objects.all():
            if re.match(mapping.source_mapping, self.path):
                yield mapping

    def _get_or_create_url_mapping(self, regexp_rule):
        target_url = generate_url(regexp_rule.source_mapping, regexp_rule.target_mapping, self.path)
        rule, _ = models.RegexpGeneratedMapping.objects.get_or_create(
            source_url=self.path, defaults={
                'regexp': regexp_rule,
                'target_url': target_url,
                'expire_after': regexp_rule.mapping_duration,
            })
        return rule

    @staticmethod
    def _redirect_response(redirect_to):
        return http.HttpResponsePermanentRedirect(redirect_to=redirect_to)

    @transaction.atomic
    def _update_usage(self, rule):
        rule = rule.lock()
        data = self._get_usage_data_dict()
        if rule.last_usage:
            rule.last_usage.delete()
        rule.last_usage = self._log_usage(data)
        rule.save()

    @staticmethod
    def _log_usage(data):
        return models.LastUsageLog.objects.create(**data)

    def _get_usage_data_dict(self):
        return {
            'used_date': timezone.now(),
            'referer': self.meta.get('HTTP_REFERER', ''),
            'user_agent': self.meta.get('HTTP_USER_AGENT', ''),
            'ip': self.meta.get('HTTP_X_REAL_IP', '0.0.0.0'),
            'user': self._get_user(),
        }

    def _get_user(self):
        if self.request.user.is_authenticated():
            return self.request.user
        else:
            return None


def generate_url(source_mapping, target_mapping, path):
    return re.sub(source_mapping, target_mapping, path)


class InvalidRedirect(Exception):
    pass



from django.core.management.base import BaseCommand
from django.utils import timezone

from url_migration import models


class Command(BaseCommand):
    def handle(self, **options):
        for rule in models.UrlRegexpMapping.objects.filter(last_usage__isnull=False):
            self._remove_if_unused(rule)
        for rule in models.UrlMapping.objects.filter(last_usage__isnull=False):
            self._remove_if_unused(rule)

    def _remove_if_unused(self, rule):
        if rule.last_usage.used_date + rule.expire_after < timezone.now():
            self.stdout.write('Removing expired rule %s' % str(rule))
            rule.delete()



from os import path
from setuptools import setup, find_packages


def read(name):
    return open(path.join(path.dirname(__file__), name)).read()

setup(
    name='django-url-migration',
    version='0.2.0',
    description='Advanced url migration application',
    long_description=read('README.rst'),
    author='Social WiFi',
    author_email='kontakt@socialwifi.com',
    url='https://github.com/socialwifi/django-url-migration',
    packages=find_packages(exclude=["url_migration_demo"]),
    install_requires=[
        'Django',
    ],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: BSD License",
        "Development Status :: 4 - Beta",
        "Operating System :: OS Independent",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2017-04-27 21:14
from __future__ import unicode_literals

import datetime
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='LastUsageLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('used_date', models.DateTimeField()),
                ('referer', models.TextField()),
                ('user_agent', models.TextField()),
                ('ip', models.GenericIPAddressField()),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UrlMapping',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('source_url', models.CharField(max_length=200, unique=True)),
                ('target_url', models.CharField(max_length=200)),
                ('expire_after', models.DurationField(default=datetime.timedelta(180), help_text='Time of inactivity after which the mapping expires. Format: "DD HH:MM:SS"')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UrlRegexpMapping',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('source_mapping', models.CharField(max_length=200, unique=True)),
                ('target_mapping', models.CharField(max_length=200)),
                ('expire_after', models.DurationField(default=datetime.timedelta(180), help_text='Time of inactivity after which the mapping expires. Format: "DD HH:MM:SS"')),
                ('mapping_duration', models.DurationField(default=datetime.timedelta(180), help_text='Time of inactivity after which the created url mapping expires. Format: "DD HH:MM:SS"')),
                ('test_source_url', models.CharField(help_text='<b><span style="color: red;">You are in the debug mode - url mapping is disabled. You can test it here only.</span></b>', max_length=200)),
                ('test_target_url', models.CharField(help_text='Provide example old/new urls you are attempting to map. Test urls will be used to test your regexp mappings.<br>For example <b>/a/(?P&lt;pk&gt;[0-9]+)/</b> mapping with <b>/b/\\g&lt;pk&gt;/</b> target mapping will change /a/1/ to /b/1/.', max_length=200)),
                ('last_usage', models.OneToOneField(editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to='url_migration.LastUsageLog')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RegexpGeneratedMapping',
            fields=[
                ('urlmapping_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='url_migration.UrlMapping')),
                ('regexp', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='url_migration.UrlRegexpMapping')),
            ],
            options={
                'abstract': False,
            },
            bases=('url_migration.urlmapping',),
        ),
        migrations.AddField(
            model_name='urlmapping',
            name='last_usage',
            field=models.OneToOneField(editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to='url_migration.LastUsageLog'),
        ),
    ]



from unittest import mock

from django.core import management
from django import test
from django.utils import timezone
import pytz

from .. import models
from . import factories


@mock.patch('django.utils.timezone.now', mock.Mock(return_value=timezone.datetime(2000, 12, 10, 22, 11, tzinfo=pytz.utc)))
class TestRemoveExpiredRedirects(test.TestCase):
    def test_if_url_mapping_is_removed(self):
        log = factories.LastUsageLogFactory(used_date=timezone.datetime(1999, 12, 10, 22, 11, tzinfo=pytz.utc))
        factories.UrlMappingFactory(last_usage=log, id=199)
        management.call_command('remove_expired_redirects')
        self.assertFalse(models.UrlMapping.objects.filter(pk=199).exists())

    def test_if_regexp_mapping_is_removed(self):
        log = factories.LastUsageLogFactory(used_date=timezone.datetime(1999, 12, 10, 22, 11, tzinfo=pytz.utc))
        factories.UrlRegexpMappingFactory(last_usage=log, id=233)
        management.call_command('remove_expired_redirects')
        self.assertFalse(models.UrlRegexpMapping.objects.filter(pk=233).exists())

    def test_if_used_mapping_is_not_removed(self):
        log = factories.LastUsageLogFactory(used_date=timezone.datetime(2001, 12, 10, 22, 11, tzinfo=pytz.utc))
        factories.UrlMappingFactory(last_usage=log, id=344)
        management.call_command('remove_expired_redirects')
        self.assertTrue(models.UrlMapping.objects.filter(pk=344).exists())

    def test_if_used_regexp_mapping_is_not_removed(self):
        log = factories.LastUsageLogFactory(used_date=timezone.datetime(2001, 12, 10, 22, 11, tzinfo=pytz.utc))
        factories.UrlRegexpMappingFactory(last_usage=log, id=422)
        management.call_command('remove_expired_redirects')
        self.assertTrue(models.UrlRegexpMapping.objects.filter(pk=422).exists())

    def test_if_unused_mapping_is_not_removed(self):
        factories.UrlMappingFactory(id=566)
        management.call_command('remove_expired_redirects')
        self.assertTrue(models.UrlMapping.objects.filter(pk=566).exists())

    def test_if_unused_regexp_mapping_is_not_removed(self):
        factories.UrlRegexpMappingFactory(id=229)
        management.call_command('remove_expired_redirects')
        self.assertTrue(models.UrlRegexpMapping.objects.filter(pk=229).exists())

    def test_if_removing_regexp_mapping_does_not_remove_generated_mappings(self):
        log = factories.LastUsageLogFactory(used_date=timezone.datetime(1999, 12, 10, 22, 11, tzinfo=pytz.utc))
        mapping = factories.UrlRegexpMappingFactory(last_usage=log, id=988)
        factories.RegexpGeneratedMappingFactory(regexp=mapping, id=1022)
        management.call_command('remove_expired_redirects')
        self.assertTrue(models.RegexpGeneratedMapping.objects.filter(pk=1022).exists())



from datetime import timedelta

import factory

from django.utils import timezone

from .. import models


class UrlMappingFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.UrlMapping

    expire_after = timedelta(days=10)


class UrlRegexpMappingFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.UrlRegexpMapping

    expire_after = timedelta(days=10)
    mapping_duration = timedelta(days=7)


class RegexpGeneratedMappingFactory(UrlMappingFactory):
    class Meta:
        model = models.RegexpGeneratedMapping

    regexp = factory.SubFactory(UrlRegexpMappingFactory)


class LastUsageLogFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.LastUsageLog

    used_date = timezone.now()
    referer = 'http://www.google.pl'
    user_agent = 'Firefox'
    ip = '10.0.0.2'
